class IdError(Exception): pass


class NullError(IdError): pass


class CharacterError(IdError): pass


class CheckDigitError(IdError): pass


class CheckSumError(IdError): pass


class CountryCodeError(IdError): pass


class LengthError(IdError): pass


class NullError(IdError): pass



from abc import ABCMeta, abstractmethod
from math import isnan

from cymbology.exceptions import (
    IdError, NullError, LengthError,
    CharacterError, CheckDigitError, CheckSumError
)


class SecurityId(metaclass=ABCMeta):
    """A financial security id that can be validated.

    Attributes
    ----------
    MIN_LEN : int
        minimum length of security id with check digit.
    MAX_LEN : int
        maximum length of security id with check digit.

    Notes
    -----
    "sid" input variable name implies security id with check digit appended.
    "sid_" input variable name implies security id w/o check digit.
    """

    MIN_LEN = 1
    MAX_LEN = None

    def validate(self, sid):
        """validate security id string.

        returns sid if is validate, else raises an IdError exception.
        """

        null_check(sid)
        check_sum = self.calculate_checksum(sid[:-1])
        check_digit = val_check_digit(sid)

        if check_sum == check_digit:
            return sid
        else:
            raise CheckSumError

    def is_valid(self, sid):
        """True if sid is valid security id string, else False."""

        try:
            return bool(self.validate(sid))
        except IdError:
            return False

    def calculate_checksum(self, sid_):
        """calculate the check digit."""

        self._id_check(sid_, offset=1)

        try:
            return self._calculate_checksum(sid_)
        except KeyError:
            raise CharacterError

    @abstractmethod
    def _calculate_checksum(self, sid_):
        NotImplementedError

    def append_checksum(self, sid_):
        """calculate and append check sum digit to security id."""

        sid_ += str(self.calculate_checksum(sid_))
        return sid_

    def __str__(self):
        return "<cymbology %s>" % self.__class__.__name__

    def _id_check(self, sid_, offset=0):

        null_check(sid_)

        if not (self.MIN_LEN - offset) <= len(sid_) <= (self.MAX_LEN - offset):
            raise LengthError

        self._additional_checks(sid_)

    def _additional_checks(self, sid_):
        pass


def null_check(sid):
    """Check if id string is null."""

    if sid is None or sid is "" or (isinstance(sid, float) and isnan(sid)):
        raise NullError


def val_check_digit(sid):
    """checks if check digit can convert to integer"""

    try:
        return int(sid[-1])
    except ValueError:
        raise CheckDigitError



from cymbology.alphanum import CHAR_MAP


def luhn_modn_checksum(sid):
    """calculate the luhn modolo n check sum."""

    return _luhnify(CHAR_MAP[c] for c in reversed(sid))


def _luhnify(gen):
    """calculates luhn sum given a generator of integers in reverse order."""

    sum_ = 0

    for index, val in enumerate(gen, 1):
        if index % 2:
            val *= 2
            sum_ += val // 10 + val % 10
        else:
            sum_ += val

    return (10 - sum_ % 10) % 10



#!/usr/bin/env python

import os
from os.path import exists
from setuptools import setup

ENCODING = 'utf-8'
PACKAGE_NAME = 'cymbology'

DESCRIPTION = 'financial identifier validation.'

local_directory = os.path.abspath(os.path.dirname(__file__))
version_path = os.path.join(local_directory, PACKAGE_NAME, '_version.py')

version_ns = {}
with open(version_path, 'r', encoding=ENCODING) as f:
    exec(f.read(), {}, version_ns)


def get_requirements(requirement_file):
    requirements = list(
        open(requirement_file, 'r',
             encoding=ENCODING).read().strip().split('\r\n'))
    return requirements


def get_read_me(file_name):
    return open(file_name).read() if exists(file_name) else DESCRIPTION


setup(
    name=PACKAGE_NAME,
    packages=[PACKAGE_NAME],
    description=DESCRIPTION,
    package_data={
        '': ['*.txt', '*.json'],
    },
    license='BSD',
    include_package_data=True,
    version=version_ns['__version__'],
    url='https://github.com/pmart123/cymbology',
    author='Philip Martin',
    author_email='philip.martin2007@gmail.com',
    extras_require={
        'develop': get_requirements('requirements-dev.txt'),
        'test': get_requirements('requirements-test.txt')
    },
    keywords='finance symbology securities entities entity-resolution',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    long_description=get_read_me('README.md'),
    zip_safe=False
)
